-----------------------------------------------------------------------------------------------
-- Client Lua Script for SimpleBuffBar
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- SimpleBuffBar Module Definition
------------------------------
local SimpleBuffBar = {} 
local cassPkg = nil
local loadConditions = {
	onDocLoaded = false,
	onPlayerUnitLoaded = false
}

SimpleBuffBar.CodeEnumUnit = {
	Player = 1,
	Target = 2,
	Focus = 3
}
SimpleBuffBar.BarConstants = {
	["PlayerBuffBar"] = {
		color = "xkcdLighterGreen",
		unit = SimpleBuffBar.CodeEnumUnit.Player,
		stockFrame = "luaUnitFrame",
		stockControl = "BeneBuffBar"
	}, 
	["PlayerDebuffBar"] =  {
		color = "xkcdLightRed",
		unit = SimpleBuffBar.CodeEnumUnit.Player,
		stockFrame = "luaUnitFrame",
		stockControl = "HarmBuffBar"
	},
	["TargetBuffBar"] = {
		color = "yellow",
		unit = SimpleBuffBar.CodeEnumUnit.Target,
		stockFrame = "luaTargetFrame",
		stockControl = "BeneBuffBar"
	}, 
	["TargetDebuffBar"] =  {
		color = "xkcdLightOrange",
		unit = SimpleBuffBar.CodeEnumUnit.Target,
		stockFrame = "luaTargetFrame",
		stockControl = "HarmBuffBar"
	},
	["FocusBuffBar"] = {
		color = "xkcdBrightSkyBlue",
		unit = SimpleBuffBar.CodeEnumUnit.Focus,
		stockFrame = "luaFocusFrame",
		stockControl = "BeneBuffBar"
	}, 
	["FocusDebuffBar"] =  {
		color = "xkcdBabyPurple",
		unit = SimpleBuffBar.CodeEnumUnit.Focus,
		stockFrame = "luaFocusFrame",
		stockControl = "HarmBuffBar"
	}
}
local barSettings = {
	scale = 1.0,
	opacity = 1.0,
	rotation = 0,
	hide = false,
	lock = false,
	hideStock = true,
	needsLoading = true,
	wnd = nil,
	unit = nil,
	location = nil
}
local userSettings = {
	disable = false,
	bars = {
		["PlayerBuffBar"] = {}, 
		["PlayerDebuffBar"] =  {},
		["TargetBuffBar"] = {}, 
		["TargetDebuffBar"] =  {},
		["FocusBuffBar"] = {}, 
		["FocusDebuffBar"] =  {}
	}
}

-- without deeepcopy which is in a package (can't use yet), we manually copy these over
for _,bar in pairs(userSettings.bars) do
	for k,v in pairs(barSettings) do
		bar[k] = v
	end
end
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function SimpleBuffBar:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function SimpleBuffBar:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "SimpleBuffBar"
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- SimpleBuffBar OnLoad
-----------------------------------------------------------------------------------------------
function SimpleBuffBar:OnLoad()
	cassPkg = Apollo.GetPackage("CassPkg-1.0").tPackage
	self.xmlDoc = XmlDoc.CreateFromFile("SimpleBuffBar.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	self.units = {}
	self.bars = {}
end

-----------------------------------------------------------------------------------------------
-- SimpleBuffBar OnDocLoaded
-----------------------------------------------------------------------------------------------
function SimpleBuffBar:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.buffWndRelocated = false
		self.deBuffWndRelocated = false
	
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "SimpleBuffBarForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		self.wndList = 	Apollo.LoadForm(self.xmlDoc, "SimpleBuffBarList", self.wndMain:FindChild("WindowScrollContainer"), self)
	    self.wndMain:Show(false, true)
		self.playerUnit = nil

		
		self:LoadBuffWindows() -- this will load the windows without a location
		Apollo.RegisterTimerHandler("InitializeTimer", "Initialize", self)
		Apollo.CreateTimer("InitializeTimer",0.25,false)
				
		Apollo.RegisterEventHandler("ChangeWorld", "OnChangeWorld", self)
		Apollo.RegisterEventHandler("CharacterCreated", "OnChangeWorld", self)
		Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnitChanged", self)
		Apollo.RegisterEventHandler("AlternateTargetUnitChanged", "OnFocusUnitChanged", self)
	
		Apollo.CreateTimer("HideShowCarbineFrames",0.25,false)
		Apollo.RegisterTimerHandler("HideShowCarbineFrames", "UpdateCarbineFrames", self)
		Apollo.StartTimer("HideShowCarbineFrames")

		Apollo.RegisterSlashCommand("buffbar", "OnSimpleBuffBarOn", self)
		Apollo.RegisterSlashCommand("sbb", "OnSimpleBuffBarOn", self)
		
		loadConditions.onDocLoaded = true
	end
end

function SimpleBuffBar:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	for _,bar in ipairs(userSettings.bars) do 
		bar.needsLoading = true
		bar.unit = nil
		bar.wnd = nil
	end
	return userSettings	
end
function SimpleBuffBar:OnRestore(eType, tSave)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	userSettings.disable = tSave.disable
	if tSave.bars ~= nil then
		userSettings.bars = cassPkg.deepcopy(tSave.bars)
	end
    for name,bar in pairs(userSettings.bars) do 
	    for k,v in pairs(barSettings) do 
			if bar[k] == nil then
				bar[k] = v
			end
		end
	end	
	
	-- from a prior version, if the old buffScale exists, convert it to the new setting
	if tSave.buffScale ~= nil then
		userSettings.bars["PlayerBuffBar"].scale = tSave.buffScale;
	end
	if tSave.deBuffScale ~= nil then
		userSettings.bars["PlayerDebuffBar"].scale = tSave.deBuffScale;
	end
	if tSave.tBuffFrameLocation ~= nil then
		userSettings.bars["PlayerBuffBar"].location = tSave.tBuffFrameLocation;
	end
	if tSave.tDeBuffFrameLocation ~= nil then
		userSettings.bars["PlayerDebuffBar"].location = tSave.tDeBuffFrameLocation;
	end

	Apollo.StopTimer("InitializeTimer")
	Apollo.StartTimer("InitializeTimer")
end

function SimpleBuffBar:Initialize()
	self:LoadUnits()
	for k,v in pairs(loadConditions) do
		if not v then
			Apollo.StartTimer("InitializeTimer")
			return
		end
	end
	self:LoadBuffWindows() -- the windows should have been loaded, this should just move them according to the restore settings.
	self:UpdateBuffBarUnits()
end

function SimpleBuffBar:OnConfigure()
	self:OnSimpleBuffBarOn()
end

-----------------------------------------------------------------------------------------------
-- SimpleBuffBar Functions
-----------------------------------------------------------------------------------------------

function SimpleBuffBar:RefreshSettings()
	for bar,_ in pairs(userSettings.bars) do
		self:SetSliderValue(bar .. "scale",userSettings.bars[bar].scale, true, 0)
		self:SetSliderValue(bar .. "opacity",userSettings.bars[bar].opacity, true, 0)
		self:SetSliderValue(bar .. "rotation",userSettings.bars[bar].rotation, false, 0)
		self.wndMain:FindChild(bar .. "hide"):SetCheck(userSettings.bars[bar].hide)
		self.wndMain:FindChild(bar .. "lock"):SetCheck(userSettings.bars[bar].lock)
		self.wndMain:FindChild(bar .. "hideStock"):SetCheck(userSettings.bars[bar].hideStock)
	end
end
-- on SlashCommand "/buffbar"
function SimpleBuffBar:OnSimpleBuffBarOn()
	self:RefreshSettings()
	for name,bar in pairs(userSettings.bars) do 
		bar.wnd:SetBGColor(ApolloColor.new(SimpleBuffBar.BarConstants[name].color))
	end
	self.wndList:Show(true,true) -- show the window
	self.wndMain:Show(true,true) -- show the window
end

function SimpleBuffBar:OnChangeWorld()
	self.playerUnit = nil
	loadConditions.onPlayerUnitLoaded = false
	Apollo.StopTimer("InitializeTimer")
	Apollo.StartTimer("InitializeTimer")
end
function SimpleBuffBar:LoadUnits() 
	local playerUnit = self.units[SimpleBuffBar.CodeEnumUnit.Player]
	if playerUnit == nil then
		playerUnit = GameLib.GetPlayerUnit()
	end
	if playerUnit ~= nil and not playerUnit:IsValid() then
		playerUnit = nil 
	end
	if playerUnit == nil then
		for k,_ in pairs(self.units) do
			self.units[k] = nil
		end		
		for _,bar in pairs(userSettings.bars) do
			if bar.wnd ~= nil then
				bar.wnd:SetUnit(nil)
			end		
		end
		loadConditions.onPlayerUnitLoaded = false
		return false
	end

	self.units[SimpleBuffBar.CodeEnumUnit.Player] = playerUnit
	loadConditions.onPlayerUnitLoaded = true
	
	-- player unit is kosher, what about the rest	
	for _,k in pairs (SimpleBuffBar.CodeEnumUnit) do
		local u = self.units[k]
		if k ~= SimpleBuffBar.CodeEnumUnit.Player and (u == nil or not u:IsValid()) then
			-- see if we can reload a new one
			self.units[k] = nil
			u = nil		
			if k == SimpleBuffBar.CodeEnumUnit.Target then
				u = playerUnit:GetTarget()
			elseif k == SimpleBuffBar.CodeEnumUnit.Focus then
				u = playerUnit:GetAlternateTarget()
			end
			if u ~= nil and u:IsValid() then
				self.units[k] = u
			end
		end
	end		

end

function SimpleBuffBar:LoadBuffWindows() 
	for name,bar in pairs(userSettings.bars) do
		self:LoadBuffWindow(bar,name)
	end		
end
function SimpleBuffBar:LoadBuffWindow(bar,name) 
	if bar.wnd == nil then
		bar.wnd = Apollo.LoadForm(self.xmlDoc, name, nil, self)
		bar.wnd:Show(not bar.hide,true)
		bar.wnd:SetBGColor(ApolloColor.new("00FFFFFF"))
		self.bars[name] = bar.wnd -- support for BuffFilter
	end	
 	if bar.wnd ~= nil and bar.location ~= nil then
		bar.wnd:MoveToLocation(WindowLocation.new(bar.location))
	end
	self:UpdateBuffBar(name)
end

function SimpleBuffBar:UpdateCarbineFrame(name, stock)
	local success = false
	if stock ~= nil then
		local bar = userSettings.bars[name]
		local frame = stock[SimpleBuffBar.BarConstants[name].stockFrame]
		if frame ~= nil then
			for _,cluster in ipairs(frame.arClusterFrames) do
				local stockControl = cluster:FindChild(SimpleBuffBar.BarConstants[name].stockControl)
				if stockControl ~= nil then
					stockControl:Show(not bar.hideStock, true)
					success = true
				end
			end
		end
	end
	return success
end

function SimpleBuffBar:UpdateCarbineFrames()
	local success = false
	local stock = Apollo.GetAddon("FastTargetFrame")
	if stock == nil then
		stock = Apollo.GetAddon("TargetFrame")
	end
	if stock ~= nil then
		success = true
		for name,bar in pairs(userSettings.bars) do
			barUpdated = self:UpdateCarbineFrame(name,stock)
			if success and not barUpdated then
				success = false
			end
		end		
	end
	if not success then
		Apollo.StartTimer("HideShowCarbineFrames")
	end
end

function SimpleBuffBar:UpdateBuffBar(name)
	local bar = userSettings.bars[name]
	bar.wnd:SetScale(bar.scale)
	bar.wnd:SetOpacity(bar.opacity)
	bar.wnd:SetRotation(bar.rotation)
	bar.wnd:SetStyle("Moveable",not bar.lock)
	bar.wnd:Show(not bar.hide,true)
end

function SimpleBuffBar:UpdateBuffBarUnits()
	for name,bar in pairs(userSettings.bars) do
		if bar.wnd ~= nil then
			local u = self.units[SimpleBuffBar.BarConstants[name].unit]
			bar.wnd:SetUnit(u)
		end
	end		
end

function SimpleBuffBar:OnTargetUnitChanged(unitTarget)
	self.units[SimpleBuffBar.CodeEnumUnit.Target] = unitTarget
	self:UpdateBuffBarUnits()
end

function SimpleBuffBar:OnFocusUnitChanged(unitTarget)
	self.units[SimpleBuffBar.CodeEnumUnit.Focus] = unitTarget
	self:UpdateBuffBarUnits()
end

-----------------------------------------------------------------------------------------------
-- SimpleBuffBarForm Functions
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function SimpleBuffBar:OnOK()
	self.wndMain:Close() -- hide the window

end

-- when the Cancel button is clicked
function SimpleBuffBar:OnCancel()
	self.wndMain:Close() -- hide the window
end

function SimpleBuffBar:SetSliderValue(sliderName,value,isPercent,roundDigits)
	local scrubValue = value
	local scrubText = value
	if isPercent then
		scrubValue = cassPkg.round(value * 100, roundDigits)
		scrubText = scrubValue ..  " %"
	else
		scrubValue = cassPkg.round(value, roundDigits)
		scrubText = scrubValue
	end
	self.wndMain:FindChild("Slider_" .. sliderName):SetValue(scrubValue)
	self.wndMain:FindChild("Label_" .. sliderName .. "Value"):SetText(scrubText)
end


-- this will take a string like PlayerBuffBarscale, search the bars and return PlayerBuffBar and scale
function SimpleBuffBar:ParseBarSetting(s)
	for k,v in pairs(userSettings.bars) do
		if string.match(s,k) then
			return k,string.gsub(s,k,"")
		end
	end
	return nil,nil
end

function SimpleBuffBar:Slider_BarChanged( wndHandler, wndControl, fNewValue, fOldValue )
	local slider = string.sub(wndControl:GetName(),8) -- setting name is after SliderBar
	local bar, setting = self:ParseBarSetting(slider)
	local percent = true
	if string.match(setting,"rotation") then
		percent = false
		userSettings.bars[bar][setting] = fNewValue
	else
		userSettings.bars[bar][setting] = fNewValue / 100
	end
	self:SetSliderValue(slider,userSettings.bars[bar][setting],percent ,0)
	self:UpdateBuffBar(bar)
end

function SimpleBuffBar:Slider_MouseDown( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	local slider = string.sub(wndControl:GetName(),8) -- setting name is after Slider_
	local bar, setting = self:ParseBarSetting(slider)
	local percent = true
	local unit = 0.01
	if string.match(setting,"rotation") then
		percent = false
		unit = 1
	end
	if eMouseButton == 0 then
		userSettings.bars[bar][setting] = userSettings.bars[bar][setting]- unit 
	else
		userSettings.bars[bar][setting] = userSettings.bars[bar][setting] + unit 
	end
	self:SetSliderValue(slider,userSettings.bars[bar][setting],percent, 0)
	self:UpdateBuffBar(bar)
end

function SimpleBuffBar:OnOptionsClosed( wndHandler, wndControl )
	self.wndMain:Show(false,true)
end

function SimpleBuffBar:Button_CheckSetting( wndHandler, wndControl, eMouseButton )
	local bar, setting = self:ParseBarSetting(wndControl:GetName())
	userSettings.bars[bar][setting] = wndControl:IsChecked()
	self:UpdateBuffBar(bar)
	local stock = Apollo.GetAddon("FastTargetFrame")
	if stock == nil then
		stock = Apollo.GetAddon("TargetFrame")
	end
	self:UpdateCarbineFrame(bar,stock)
end

function SimpleBuffBar:OnOptionsWindowHide( wndHandler, wndControl )
	for _,bar in pairs(userSettings.bars) do 
		if bar.wnd ~= nil then
			bar.wnd:SetBGColor(ApolloColor.new("00FFFFFF"))
		end
	end
end

---------------------------------------------------------------------------------------------------
-- SimpleBeneBuffBar Functions
---------------------------------------------------------------------------------------------------

function SimpleBuffBar:OnGenerateBuffTooltip(wndHandler, wndControl, tType, splBuff)
	if wndHandler == wndControl then
		return
	end
	if Tooltip and Tooltip.GetBuffTooltipForm then 
		Tooltip.GetBuffTooltipForm(self, wndControl, splBuff, {bFutureSpell = false})
	end
end

---------------------------------------------------------------------------------------------------
-- PlayerBuffBar Functions
---------------------------------------------------------------------------------------------------

function SimpleBuffBar:OnBarMove( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
	local bar = userSettings.bars[wndControl:GetName()]
	bar.location = wndControl:GetLocation():ToTable()
end

-----------------------------------------------------------------------------------------------
-- SimpleBuffBar Instance
-----------------------------------------------------------------------------------------------
local SimpleBuffBarInst = SimpleBuffBar:new()
SimpleBuffBarInst:Init()
